#   Copyright 2010-2016 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Wunderground Plugin"""

import pywunderground

from pyhole.core import plugin
from pyhole.core import utils


class Wunderground(plugin.Plugin):
    """Provide access to current weather data."""

    @plugin.hook_add_command("weather")
    @utils.spawn
    def wunderground(self, message, params=None, **kwargs):
        """Display current weather report (ex: .w [set] [<location>])."""
        wunderground = utils.get_config("Wunderground")
        api_key = wunderground.get("key")

        if params:
            location = params
            if location.startswith("set "):
                location = location[4:]
                utils.write_file(self.name, message.source, location)
                message.dispatch("Location information saved.")
        else:
            location = utils.read_file(self.name, message.source)
            if not location:
                message.dispatch(self.weather.__doc__)
                return

        try:
            w = pywunderground.request(api_key, ["conditions"], location)
        except Exception:
            message.dispatch("Unable to fetch Wunderground data.")
            return

        if "current_observation" in w:
            w = w["current_observation"]

            city = w["display_location"]["full"]
            zip_code = w["display_location"]["zip"]
            temp = w["temperature_string"]
            humidity = w["relative_humidity"]
            wind = w["wind_string"]
            condition = w["weather"]

            zip_code = "" if zip_code == "00000" else " %s" % zip_code
            humidity = "N/A%" if len(humidity) > 3 else humidity

            result = "%s%s: %s   Humidity: %s   Wind: %s   %s" % (city,
                                                                  zip_code,
                                                                  temp,
                                                                  humidity,
                                                                  wind,
                                                                  condition)
            message.dispatch(result)
        else:
            message.dispatch("Location not found: '%s'" % location)

    @plugin.hook_add_command("w")
    def alias_w(self, message, params=None, **kwargs):
        """Alias of wunderground."""
        self.wunderground(message, params, **kwargs)



#   Copyright 2016 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Jira Plugin"""

import json

from pyhole.core import plugin
from pyhole.core import utils


class JiraClient(object):
    def __init__(self):
        jira = utils.get_config("Jira")

        self.auth_server = jira.get("auth_server")
        self.domain = jira.get("domain")
        self.username = jira.get("username")
        self.password = jira.get("password")

    def get(self, issue_id):
        url = "%s/rest/api/latest/issue/%s" % (self.auth_server, issue_id)

        return utils.fetch_url(url, verify=False,
                               auth=(self.username, self.password))


class Jira(plugin.Plugin):
    """Provide access to the Jira API."""

    def __init__(self, session):
        self.session = session
        self.name = self.__class__.__name__

        self.client = JiraClient()

    @plugin.hook_add_msg_regex("([A-Z]{2}-[0-9]{4,5})")
    def regex_match_issue(self, message, match, **kwargs):
        """Retrieve Jira ticket information (ex: AB-1234)."""
        try:
            issue_id = match.group(0)
            self._find_issue(message, issue_id)
        except Exception:
            return

    @utils.spawn
    def _find_issue(self, message, issue_id):
        """Find and display a Jira issue."""
        issue = json.loads(self.client.get(issue_id).content)

        msg = "%s: %s [Status: %s, Priority: %s, Assignee: %s] %s"
        message.dispatch(msg % (
            issue_id,
            issue["fields"]["summary"],
            issue["fields"]["status"]["name"],
            issue["fields"]["priority"]["name"],
            issue["fields"]["assignee"]["displayName"],
            "%s/jira/browse/%s" % (self.client.domain, issue_id)
        ))



#   Copyright 2011-2016 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Kernel.org Plugin"""

import re

from pyhole.core import plugin
from pyhole.core import utils


class Kernel(plugin.Plugin):
    """Provide access to kernel.org data."""

    @plugin.hook_add_command("kernel")
    @utils.spawn
    def kernel(self, message, params=None, **kwargs):
        """Retrieve the current kernel version (ex: .kernel)"""
        url = "https://www.kernel.org/kdist/finger_banner"
        response = utils.fetch_url(url)
        if response.status_code != 200:
            return

        r = re.compile("(.* mainline .*)")
        m = r.search(response.content)
        kernel = m.group(1).replace("  ", "")
        message.dispatch(kernel)



#   Copyright 2011-2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Log Unit Tests"""

import os
import unittest

from pyhole.core import log
from pyhole.core import utils


class TestLog(unittest.TestCase):
    def test_logger(self):
        test_log_dir = utils.get_home_directory() + "logs/"
        log.setup_logger(name="test")
        test_log = log.get_logger("TEST")
        self.assertEqual("TEST", test_log.name)
        self.assertEqual(test_log.level, 0)
        os.unlink(test_log_dir + "test.log")



#   Copyright 2011-2016 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole URL Plugin"""

from BeautifulSoup import BeautifulSoup

from pyhole.core import plugin
from pyhole.core import utils


class Url(plugin.Plugin):
    """Provide access to URL data."""

    def __init__(self, session):
        self.session = session
        self.name = self.__class__.__name__
        self.url = None

    @plugin.hook_add_command("title")
    def title(self, message, params=None, **kwargs):
        """Display the title of the a URL (ex: .title [<url>])"""
        if params:
            self.url = params.split(" ", 1)[0]

        if not _matches_host(self.url):
            self._find_title(message, self.url)

    @plugin.hook_add_msg_regex("(https?://|www.)[^\> ]+")
    def regex_match_url(self, message, match, **kwargs):
        """Watch and keep track of the latest URL."""
        try:
            # NOTE(jk0): Slack does some weird things with URLs.
            self.url = match.group(0).split("|", 1)[0]

            if _matches_host(self.url):
                self._find_title(message, self.url)
        except TypeError:
            return

    @utils.spawn
    def _find_title(self, message, url):
        """Find the title of a given URL."""
        # NOTE(jk0): Slack does some weird things with URLs.
        url = url.replace("<", "").replace(">", "").split("|")[0]
        if not url.startswith(("http://", "https://")):
            url = "http://" + url

        try:
            response = utils.fetch_url(url)
        except Exception:
            return

        soup = BeautifulSoup(response.content)
        if soup.head:
            title = utils.decode_entities(soup.head.title.string)
            content_type = response.headers.get("Content-Type")
            message.dispatch("%s (%s)" % (title, content_type))
        else:
            message.dispatch("No title found: %s" % url)


def _matches_host(url):
    """Watch for certain websites."""
    try:
        host = url.split("/")[2]
    except IndexError:
        host = url

    lookup_sites = ("open.spotify.com", "www.youtube.com", "youtu.be")
    if host.startswith(lookup_sites):
        return True



#   Copyright 2014 Philip Schwartz
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import re

from ..color import UnknownColor
from ..colormap import ColorMap


class MircColors(ColorMap):
    """Mirc Color code schema
        Defined at http://www.mirc.com/colors.html
    """

    _attribute = u'\x03'
    _separator = u','
    _digit = u'\x16\x16'
    _reset = u'\x0f'
    _clean = re.compile(r"([\x02\x0F\x1F\x0E\x16\x1B]|\x03([0-9]{0,2})(,([0-9"
                        "]{0,2}))?|\x04[0-9A-Fa-f]{6}(,([0-9A-Fa-f]){6})?)*")
    _color_map = {
        u'white': u'0',
        u'black': u'1',
        u'blue': u'2',
        u'green': u'3',
        u'red': u'4',
        u'brown': u'5',
        u'purple': u'6',
        u'orange': u'7',
        u'yellow': u'8',
        u'light green': u'9',
        u'teal': u'10',
        u'light cyan': u'11',
        u'light blue': u'12',
        u'pink': u'13',
        u'grey': u'14',
        u'light grey': u'15'
    }

    def color(self, _color):
        _color = _color.lower().strip()
        if _color not in self._color_map:
            raise (UnknownColor, _color)
        return self._color_map[_color]

    def strip(self, text):
        text = self._clean.sub(u'', text)
        text = text.replace(self._reset, u'')
        return text



#   Copyright 2010-2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Utility Library"""

from __future__ import with_statement

import eventlet
import optparse
import os
import re
import requests
import sys
import traceback

from BeautifulSoup import BeautifulStoneSoup
from functools import wraps
from multiprocessing import Process
from multiprocessing import Queue

import config
import version


eventlet.monkey_patch()


def admin(func):
    """Require admin rights."""
    def wrap(self, message, *args, **kwargs):
        if message.source in self.session.admins:
            return func(self, message, *args, **kwargs)

        return message.dispatch("Sorry, you are not authorized to do that.")

    wrap.__doc__ = func.__doc__
    wrap.__name__ = func.__name__
    wrap.__module__ = func.__module__

    return wrap


def require_params(func):
    """Require parameters."""
    def wrap(self, message, params, *args, **kwargs):
        if not params:
            message.dispatch(wrap.__doc__)
            return

        return func(self, message, params, *args, **kwargs)

    wrap.__doc__ = func.__doc__
    wrap.__name__ = func.__name__
    wrap.__module__ = func.__module__

    return wrap


def spawn(func):
    """Greenthread Spawning Decorator"""
    def wrap(self, *args, **kwargs):
        eventlet.spawn_n(func, self, *args, **kwargs)

    wrap.__doc__ = func.__doc__
    wrap.__name__ = func.__name__
    wrap.__module__ = func.__module__

    return wrap


def subprocess(func):
    """Decorator running function in subprocess."""
    def _subprocess(q, *args, **kwargs):
        try:
            ret = func(*args, **kwargs)
        except Exception:
            ex_type, ex_value, tb = sys.exc_info()
            error = ex_type, ex_value, "".join(traceback.format_tb(tb))
            ret = None
        else:
            error = None

        q.put((ret, error))

    @wraps(func)
    def wrapper(*args, **kwargs):
        q = Queue()
        p = Process(target=_subprocess, args=[q] + list(args), kwargs=kwargs)
        p.start()
        p.join()
        ret, error = q.get()

        if error:
            ex_type, ex_value, tb_str = error
            message = "%s (in subprocess)\n%s" % (ex_value.message, tb_str)
            raise ex_type(message)

        return ret

    return wrapper


def decode_entities(html):
    """Strip HTML entities from a string and make it printable"""
    html = re.sub("\n", "", html)
    html = re.sub(" +", " ", html)
    html = " ".join(str(x).strip() for x in BeautifulStoneSoup(html,
                    convertEntities=BeautifulStoneSoup.HTML_ENTITIES).findAll(
                    text=True))

    return filter(lambda x: ord(x) > 9 and ord(x) < 127, html)


def ensure_int(param):
    """Ensure the given param is an int"""
    try:
        param = re.sub("\W", "", param)

        return int(param)
    except ValueError:
        return None


def build_options():
    """Generate command line options"""
    parser = optparse.OptionParser(version=version.version_string())
    parser.add_option("-c", "--config", default=get_conf_file_path(),
                      help="specify the path to a configuration file")
    parser.add_option("-d", "--debug", action="store_true",
                      help="show debugging output")

    return parser.parse_args()


def get_option(option):
    """Retrive an option from the command line."""
    options, _args = build_options()

    return vars(options).get(option)


def get_home_directory():
    """Return the home directory"""
    home_dir = os.getenv("HOME") + "/.pyhole/"
    if not os.path.exists(home_dir):
        os.makedirs(home_dir)

    return home_dir


def get_directory(new_dir):
    """Return a directory"""
    home_dir = get_home_directory()
    new_dir = home_dir + new_dir

    if not os.path.exists(new_dir):
        os.makedirs(new_dir)

    return new_dir + "/"


def get_conf_file_path():
    """Return the path to the conf file"""
    return get_home_directory() + "pyhole.conf"


def get_conf_file():
    """Return the path to the conf file"""
    return get_option("config") or get_conf_file_path()


def get_config(section="Pyhole"):
    """Return the default config object"""
    return config.Config(get_conf_file(), section)


def write_file(directory, file_name, data):
    """Write data to file"""
    directory = get_directory(directory)
    with open(directory + file_name, "w") as open_file:
        open_file.write(str(data).strip())


def read_file(directory, file_name):
    """Read and return the data in file"""
    directory = get_directory(directory)
    try:
        with open(directory + file_name, "r") as open_file:
            data = open_file.read()

        return data
    except IOError:
        return None


def list_files(directory):
    directory = get_directory(directory)

    return os.listdir(directory)


def generate_config():
    """Generate an example config"""
    example = """# Global Configuration

[Pyhole]
admins = nick!ident, nick2!ident, slack.username
command_prefix = .
reconnect_delay = 60
rejoin_delay = 5
debug = False
plugins = admin
networks = FreeNode, EFnet, SlackNetwork

[GoogleMaps]
key = abcd1234

[Jira]
auth_server = auth.jira.example.com
domain = jira.example.com
username = abcd1234
password = pass1234

[Redmine]
domain = redmine.example.com
key = abcd1234

[Wunderground]
key = abcd1234

[XSA]
notify = #channel1, #channel2

# Network Configuration

[FreeNode]
server = verne.freenode.net
username =
password =
port = 7000
ssl = True
ipv6 = True
bind_to = fe80::1
nick = mynick
identify_password = abcd1234
channels = #mychannel key, #mychannel2

[EFnet]
server = irc.efnet.net
username =
password =
port = 6667
ssl = False
ipv6 = False
bind_to =
nick = mynick
identify_password =
channels = #mychannel key, #mychannel2

[SlackNetwork]
api_token = abcd1234
nick = mynick
"""

    conf_file = get_conf_file()
    if os.path.exists(conf_file):
        return

    print "Generating..."
    with open(conf_file, "w") as open_file:
        open_file.write(example)
    print "Done"


def fetch_url(url, **kwargs):
    """Fetch a URL."""
    session = requests.Session()
    session.headers.update({
        "User-Agent": "pyhole/%s" % version.version()
    })
    return session.get(url, **kwargs)



#   Copyright 2015-2016 Rick Harris
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Distance to User Plugin"""

import pywunderground

from pyhole.core import plugin
from pyhole.core import utils


class Distance(plugin.Plugin):
    """Display distance between two users by using their weather data."""

    @plugin.hook_add_command("distance")
    @utils.require_params
    @utils.spawn
    def distance(self, message, params=None, **kwargs):
        """Display distances (ex: .dist <nick|loc> [to <nick|loc>])."""
        maps_api = utils.get_config("GoogleMaps")
        try:
            key = maps_api.get("key")
        except Exception:
            message.dispatch("No Google Maps API key set.")
            return

        parts = params.split(" to ")
        if not parts:
            message.dispatch(self.distance.__doc__)
            return

        dest_nick = parts[0].strip()

        if len(parts) > 1:
            origin_nick = parts[1].strip()
        else:
            origin_nick = message.source.split("!")[0]

        dest = None
        origin = None
        for filename in utils.list_files("Wunderground"):
            nick = filename.split("!")[0]
            if nick == dest_nick:
                dest = utils.read_file("Wunderground", filename)
            if nick == origin_nick:
                origin = utils.read_file("Wunderground", filename)

        if not dest:
            # They passed in a location
            dest = dest_nick

        if not origin:
            # They passed in a location
            origin = origin_nick

        origin = _resolve_pws(origin)
        dest = _resolve_pws(dest)

        resp = utils.fetch_url("https://maps.googleapis.com/maps/api"
                               "/directions/json?origin=%s&destination=%s"
                               "&key=%s" % (origin, dest, key))

        msg = None
        if resp.status_code == 200:
            try:
                msg = resp.json()["routes"][0]["legs"][0]["distance"]["text"]
            except IndexError:
                pass

        if not msg:
            msg = "Unable to fetch data from Google Maps."

        message.dispatch(msg)


def _resolve_pws(location):
    """Look up the location of a PWS."""
    if location.lower().startswith("pws:"):
        wunderground = utils.get_config("Wunderground")
        api_key = wunderground.get("key")

        w = pywunderground.request(api_key, ["conditions"], location)

        return w["current_observation"]["display_location"]["zip"]

    return location



#   Copyright 2015-2016 Rick Harris
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Xen Security Advisory (XSA) Plugin"""

import json

from BeautifulSoup import BeautifulSoup

from pyhole.core import plugin
from pyhole.core import utils


class Xsa(plugin.Plugin):
    """Display information regarding Xen Security Advisories (XSAs)."""
    XSA_URL = "http://xenbits.xen.org/xsa/"

    @plugin.hook_add_keyword("xsa")
    @utils.require_params
    @utils.spawn
    def keyword_xsa(self, message, params=None, **kwargs):
        """Retrieve XSA information (ex: xsa123 or xsa-123)."""
        # abs is needed in case we get 'xsa-123', that would be -123
        xsa_num = abs(utils.ensure_int(params))
        xsa_id = "XSA-%d" % xsa_num
        data = self._load_cached_xsa_data()
        msg = self._make_xsa_message(xsa_id, data)
        if not msg:
            msg = "Unable to find matching XSA."
        message.dispatch(msg)

    def _make_xsa_message(self, xsa_id, data):
        try:
            xsa_info = data[xsa_id]
        except:
            return None
        else:
            # Use ID if we don't have a link
            if not xsa_info["link"]:
                xsa_info["link"] = xsa_info["id"]
            return ("%(link)s: %(title)s (Public Release: %(public_release)s)"
                    % xsa_info)

    @plugin.hook_add_poll("poll_xsa_list", poll_timer=300)
    def poll_xsa_list(self, message, params=None, **kwargs):
        old_data = self._load_cached_xsa_data()

        new_data = self._fetch_xsa_data()
        if not new_data:
            return

        new_data_json = json.dumps(new_data)
        utils.write_file(self.name, "xsas.json", new_data_json)

        # Notify subscribed channels on new XSAs
        old_xsa_ids = set(old_data.keys())
        new_xsa_ids = set(new_data.keys())
        added_xsa_ids = new_xsa_ids - old_xsa_ids
        if old_xsa_ids and added_xsa_ids:
            self._notify_new_xsas(added_xsa_ids, new_data)

    def _notify_new_xsas(self, added_xsa_ids, new_data):
        xsa_config = utils.get_config("XSA")
        try:
            notify_str = xsa_config.get("notify")
        except Exception:
            return

        channels = notify_str.split(",")
        for xsa_id in added_xsa_ids:
            msg = self._make_xsa_message(xsa_id, new_data)
            for channel in channels:
                channel = channel.strip()
                self.irc.notice(channel, msg)

    def _load_cached_xsa_data(self):
        data_json = utils.read_file(self.name, "xsas.json")
        if data_json is None:
            return {}
        return json.loads(data_json)

    def _fetch_xsa_data(self):
        resp = utils.fetch_url(self.XSA_URL)
        if resp.status_code != 200:
            print "Error fetching XSAs: %s" % resp.status_code
            return {}
        soup = BeautifulSoup(resp.text)
        rows = soup.findAll("tr")
        data = {}
        for row in rows:
            children = row.findChildren("td")
            if not children:
                continue
            try:
                xsa_id = children[0].contents[0].string
            except:
                continue
            try:
                xsa_link = "".join([self.XSA_URL,
                                    children[0].contents[0]["href"]])
            except:
                xsa_link = None
            try:
                public_release = children[1].contents[0]
            except:
                public_release = None
            try:
                title = children[5].contents[0]
            except:
                title = None
            data[xsa_id] = {"id": xsa_id,
                            "link": xsa_link,
                            "public_release": public_release,
                            "title": title}
        return data



#   Copyright 2010-2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Network Process Class"""

import multiprocessing
import sys
import time

from pyhole.core import log
from pyhole.core import utils


LOG = log.get_logger()
CONFIG = utils.get_config()


class Process(multiprocessing.Process):
    """A network connection process."""
    def __init__(self, network):
        super(Process, self).__init__()
        self.network = network
        self.reconnect_delay = CONFIG.get("reconnect_delay", type="int")

    def run(self):
        """A network connection."""
        while True:
            try:
                network_config = utils.get_config(self.network)
                if network_config.get("api_token", default=None):
                    from pyhole.core.slack import client
                else:
                    from pyhole.core.irc import client

                connection = client.Client(self.network)
            except Exception, exc:
                LOG.exception(exc)
                LOG.error("Retrying in %d seconds" % self.reconnect_delay)

                time.sleep(self.reconnect_delay)

                continue

            try:
                connection.start()
            except KeyboardInterrupt:
                sys.exit(0)
            except Exception, ex:
                LOG.exception(ex)
                LOG.error("Retrying in %d seconds" % self.reconnect_delay)

                time.sleep(self.reconnect_delay)

                continue



#   Copyright 2014 Philip Schwartz
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


class UnknownColor(Exception):
    """Raised when an unknown color is requested"""


class Color(object):
    """API for accessing colorization in a generic manner"""

    def __init__(self, _colors=None):
        if not _colors:
            from colormaps.mirccolors import MircColors
            _colors = MircColors
        self._colors = _colors()

    def colorize(self, fg=None, bg=None, text=None):
        """Generate colored text using passed parameters"""
        if not fg and not bg:
            return self._colors._reset
        if fg:
            fg = self._colors.color(fg)
        if bg:
            bg = self._colors.color(bg)
        codes = []
        if fg is None:
            codes.append(u'')
        else:
            codes.append(fg)
        if bg is not None:
            codes.append(bg)
        color = self._colors._attribute + self._colors._separator.join(codes)
        if text is not None and text[0].isdigit():
            color += self._colors._digit
        if text is not None:
            return u'%s%s%s' % (color, text, self._colors._reset)
        else:
            return color

    # added with the api into the ColorMap to strip color codes if needed.
    def clean(self, text):
        """Strip color codes from passed string."""
        return self._colors.strip(text)



#   Copyright 2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Version Unit Tests"""

import unittest

from pyhole.core import version


class TestVersion(unittest.TestCase):
    def test_current_git_hash(self):
        self.assertEqual(len(version.current_git_hash()), 5)

    def test_version_string(self):
        self.assertTrue(version.version_string().startswith("pyhole v"))

    def test_version_hash(self):
        self.assertEqual(len(version.version_hash()), 12)

    def test_version(self):
        self.assertEqual(len(version.version()), 6)



#   Copyright 2014-2015 Philip Schwartz
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Color Unit Tests"""

import re
import unittest

from pyhole.core import color
from pyhole.core import colormap


class MockColors(colormap.ColorMap):
    _attribute = u'\x03'
    _separator = u','
    _digit = u'\x16\x16'
    _reset = u'\x0f'
    _clean = re.compile(r"([\x02\x0F\x1F\x0E\x16\x1B]|\x03([0-9]{0,2})(,([0-9]"
                        "{0,2}))?|\x04[0-9A-Fa-f]{6}(,([0-9A-Fa-f]){6})?)*")
    _color_map = {
        u'red': u'0',
        u'white': u'1',
        u'blue': u'2'
    }

    def color(self, _color):
        _color = _color.lower().strip()
        if _color not in self._color_map:
            raise color.UnknownColor, _color
        return self._color_map[_color]

    def strip(self, text):
        text = self._clean.sub(u'', text)
        text = text.replace(self._reset, u'')
        return text


class TestColor(unittest.TestCase):
    def setUp(self):
        self._colors = color.Color(MockColors)

    def test_colorize(self):
        result = self._colors.colorize("red", "blue", "test")
        self.assertEqual(result, u'\x030,2test\x0f')

    def test_colorize_fg(self):
        result = self._colors.colorize("red", text="test")
        self.assertEqual(result, u'\x030test\x0f')

    def test_colorize_bg(self):
        result = self._colors.colorize(bg="blue", text="test")
        self.assertEqual(result, u'\x03,2test\x0f')

    def test_long_colorize(self):
        enc = self._colors.colorize("red", text="test")
        result = enc + " Nice " + enc
        self.assertEqual(result, u'\x030test\x0f Nice \x030test\x0f')

    def test_clean(self):
        result = self._colors.clean(u'\x030,2test\x0f')
        self.assertEqual(result, u'test')

    def test_long_clean(self):
        result = self._colors.clean(u'\x030test\x0f Nice \x030test\x0f')
        self.assertEqual(result, u'test Nice test')



#   Copyright 2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Utils Unit Tests"""

import os
import unittest

from pyhole.core import utils


class TestUtils(unittest.TestCase):
    def setUp(self):
        utils.write_file("tests", "pyhole_test_file", "foo")
        self.new_file = utils.get_home_directory() + "tests/pyhole_test_file"
        self.cpid = os.getpid()

    def tearDown(self):
        os.unlink(self.new_file)
        os.rmdir(self.new_file[:-16])

    def test_decode_entities(self):
        test_str = "<foo>&#64;&amp;bar&amp;&#64;</foo>"
        self.assertEqual(utils.decode_entities(test_str), "@&bar&@")

    def test_ensure_int(self):
        self.assertEqual(utils.ensure_int("3"), 3)

    def test_ensure_int_2(self):
        self.assertEqual(utils.ensure_int("\W"), None)

    def test_ensure_int_3(self):
        self.assertEqual(utils.ensure_int("a"), None)

    def test_build_options(self):
        options, _args = utils.build_options()
        self.assertTrue(isinstance(options, object))

    def test_get_option(self):
        conf_file = utils.get_option("config")
        self.assertTrue(conf_file.endswith("pyhole.conf"))

    def test_get_home_directory(self):
        self.assertTrue(utils.get_home_directory().endswith("/.pyhole/"))

    def test_get_directory(self):
        new_dir = utils.get_home_directory() + "test_dir"
        self.assertFalse(os.path.exists(new_dir))
        utils.get_directory("test_dir")
        self.assertTrue(os.path.exists(new_dir))
        os.rmdir(new_dir)

    def test_get_conf_file_path(self):
        conf_file_path = utils.get_conf_file_path()
        actual_conf_file_path = utils.get_home_directory() + "pyhole.conf"
        self.assertEqual(conf_file_path, actual_conf_file_path)

    def test_get_conf_file(self):
        actual_conf_file = utils.get_conf_file_path()
        generated_conf_file = utils.get_conf_file()
        self.assertEqual(actual_conf_file, generated_conf_file)

    def test_get_config(self):
        config = utils.get_config()
        debug = config.get("debug", type="bool")
        self.assertTrue(debug in [True, False])

    def test_write_file(self):
        self.assertTrue(os.path.exists(self.new_file))

    def test_read_file(self):
        self.assertEquals(utils.read_file("tests", "pyhole_test_file"), "foo")

    def test_subprocess(self):
        @utils.subprocess
        def test_pid():
            return os.getpid()

        self.assertNotEquals(self.cpid, test_pid())



#   Copyright 2015-2016 Jason Meridth
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Stocks Plugin"""

import yahoo_finance

from pyhole.core import plugin
from pyhole.core import utils


class Stocks(plugin.Plugin):
    """Provide access to current stock values."""

    @plugin.hook_add_command("stock")
    @utils.require_params
    @utils.spawn
    def stocks(self, message, params=None, **kwargs):
        """Display current stock values (ex: .stock rax,yhoo,aapl)."""
        try:
            stocks = []
            symbols = params.upper().split(",")
            for symbol in symbols:
                share = yahoo_finance.Share(symbol)
                stocks.append("%s: %s (%s)" % (symbol, share.get_price(),
                                               share.get_change()))

            message.dispatch(", ".join(stocks))
        except Exception:
            message.dispatch("Unable to fetch stock data.")



#   Copyright 2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Slack Message Class"""


class Message(object):
    def __init__(self, session, message):
        self.session = session
        self.message = message

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, _message):
        self._message = _message

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, _session):
        self._session = _session

    def dispatch(self, reply):
        raise NotImplementedError("Message Dispatcher is not implemented")

    @staticmethod
    def getMessage(**kwargs):
        return kwargs.pop("full_message")


class Reply(Message):
    def __init__(self, session, message, source, target):
        super(Reply, self).__init__(session, message)
        self.source = source
        self.target = target

    def dispatch(self, msg):
        """Dispatch message as a reply."""
        if self.session.addressed:
            self.session.reply(self.target, "%s: %s" % (self.source, msg))
            self.session.log.info("-%s- <%s> %s: %s" % (self.target,
                                                        self.session.nick,
                                                        source, msg))
        else:
            self.session.reply(self.target, msg)
            self.session.log.info("-%s- <%s> %s" % (self.target,
                                                    self.session.nick,
                                                    msg))



#   Copyright 2014 Philip Schwartz
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from abc import ABCMeta, abstractmethod, abstractproperty


class ColorMap:
    __metaclass__ = ABCMeta

    @abstractproperty
    def _attribute(self):
        pass

    @abstractproperty
    def _separator(self):
        pass

    @abstractproperty
    def _digit(self):
        pass

    @abstractproperty
    def _color_map(self):
        pass

    @abstractproperty
    def _reset(self):
        pass

    @abstractproperty
    def _clean(self):
        pass

    @abstractmethod
    def color(self):
        pass

    @abstractmethod
    def strip(self):
        pass



#   Copyright 2016 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole On-Call Plugin"""

from pyhole.core import plugin
from pyhole.core import utils


class OnCall(plugin.Plugin):
    """Manage on-call duties."""

    def __init__(self, session):
        self.session = session
        self.name = self.__class__.__name__
        self.level = "GREEN"

    @plugin.hook_add_command("note")
    @utils.require_params
    @utils.spawn
    def note(self, message, params=None, **kwargs):
        """Manage notes during an outage (ex: .note <message>)."""
        return

    @plugin.hook_add_command("status")
    @utils.require_params
    @utils.spawn
    def status(self, message, params=None, **kwargs):
        """Manage status of a service (ex: .status <service> [<status>])."""
        return

    @plugin.hook_add_command("threat")
    def threat(self, message, params=None, **kwargs):
        """Set the current threat level (ex: .threat <level>)."""
        levels = ("RED", "ORANGE", "YELLOW", "BLUE", "GREEN")

        if params:
            self.level = params.split(" ", 1)[0].upper()
            if self.level in levels:
                message.dispatch("Threat Level: %s" % self.level)
            else:
                message.dispatch("Available Levels: %s" % ", ".join(levels))
        else:
            message.dispatch("Threat Level: %s" % self.level)
            return



#   Copyright 2010-2016 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Search Plugin"""

import json
import re

from BeautifulSoup import BeautifulSoup

from pyhole.core import plugin
from pyhole.core import utils


class Search(plugin.Plugin):
    """Provide access to search engines."""

    @plugin.hook_add_command("google")
    @utils.require_params
    @utils.spawn
    def google(self, message, params=None, **kwargs):
        """Search Google (ex: .google <query>)."""
        url = "http://ajax.googleapis.com/ajax/services/search/web?v=1.0"
        response = utils.fetch_url(url, params={"q": params})
        if response.status_code != 200:
            return

        json_obj = json.loads(response.content)
        results = json_obj["responseData"]["results"]
        if results:
            for r in results:
                message.dispatch("%s: %s" % (
                                 r["titleNoFormatting"]
                                 .encode("ascii", "ignore"),
                                 r["unescapedUrl"]))
        else:
            message.dispatch("No results found: '%s'" % params)

    @plugin.hook_add_command("urban")
    @utils.require_params
    @utils.spawn
    def urban(self, message, params=None, **kwargs):
        """Search Urban Dictionary (ex: .urban <query>)."""
        url = "http://www.urbandictionary.com/define.php"
        response = utils.fetch_url(url, params={"term": params})
        if response.status_code != 200:
            return

        soup = BeautifulSoup(response.content)

        try:
            meaning = soup.find("div", {"class": "meaning"}).text
            example = soup.find("div", {"class": "example"}).text
        except AttributeError:
            message.dispatch("No results found: '%s'" % params)

        meaning = utils.decode_entities(meaning)
        example = utils.decode_entities(example)

        message.dispatch("%s (ex: %s)" % (meaning, example))

    @plugin.hook_add_command("wikipedia")
    @utils.require_params
    @utils.spawn
    def wikipedia(self, message, params=None, **kwargs):
        """Search Wikipedia (ex: .wikipedia <query>)."""
        url = "https://en.wikipedia.org/w/api.php"
        response = utils.fetch_url(url, params={
            "action": "query",
            "generator": "allpages",
            "gaplimit": 4,
            "gapfrom": params,
            "format": "json"
        })

        if response.status_code != 200:
            return

        pages = json.loads(response.content)["query"]["pages"]
        for page in pages.values():
            title = page["title"]
            title = re.sub(" ", "_", title)
            message.dispatch("http://en.wikipedia.org/wiki/%s" % title)



#   Copyright 2010-2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""pyhole - A modular IRC & Slack bot for Python developers."""

import sys
import time

from pyhole.core import log
from pyhole.core import process
from pyhole.core import utils
from pyhole.core import version


LOG = log.get_logger()
CONFIG = utils.get_config()


def Main():
    """Main IRC loop."""
    networks = CONFIG.get("networks", type="list")
    log.setup_logger()
    LOG.info("Starting %s" % version.version_string())
    LOG.info("Connecting to networks: %s" % ", ".join(networks))

    procs = []
    for network in networks:
        proc = process.Process(network)
        proc.start()
        procs.append(proc)

    try:
        while True:
            time.sleep(1)
            for proc in procs:
                if not proc.is_alive():
                    procs.remove(proc)

            if not procs:
                LOG.info("No longer connected to any networks, shutting down")
                sys.exit(0)
    except KeyboardInterrupt:
        LOG.info("Caught KeyboardInterrupt, shutting down")



#   Copyright 2013 Philip Schwartz
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Message"""

import irc.client as irclib


class Message(object):
    def __init__(self, session, message):
        self.session = session
        self.message = message

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, _message):
        self._message = _message

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, _session):
        self._session = _session

    def _mangle_msg(self, msg):
        """Prepare the message for sending."""
        if not hasattr(msg, "encode"):
            try:
                msg = str(msg)
            except Exception:
                self.session.log.error("msg cannot be converted to string")
                return

        msg = msg.split("\n")
        # NOTE(jk0): 10 is completely arbitrary for now.
        if len(msg) > 10:
            msg = msg[0:8]
            msg.append("...")

        return msg

    def dispatch(self, reply):
        raise NotImplementedError("Message Dispatcher is not implemented")

    @staticmethod
    def getMessage(**kwargs):
        return kwargs.pop("full_message")


class Notice(Message):
    def __init__(self, session, message, target):
        super(Notice, self).__init__(session, message)
        self.target = target

    def dispatch(self, reply):
        """Dispatch message as notice."""
        _reply = self._mangle_msg(reply)
        for line in _reply:
            self.session.notice(self.target, line)
            if irclib.is_channel(self.target):
                self.session.log.info("-%s- <%s> %s" % (self.target,
                                                        self.session.nick,
                                                        line))
            else:
                self.session.log.info("<%s> %s" % (self.session.nick, line))


class Reply(Message):
    def __init__(self, session, message, source, target):
        super(Reply, self).__init__(session, message)
        self.source = source
        self.target = target

    def dispatch(self, reply):
        """dispatch message as a reply."""
        _reply = self._mangle_msg(reply)
        for line in _reply:
            if self.session.addressed:
                source = self.source.split("!")[0]
                self.session.privmsg(self.target, "%s: %s" % (source, line))
                self.session.log.info("-%s- <%s> %s: %s" % (self.target,
                                                            self.session.nick,
                                                            source, line))
            else:
                self.session.privmsg(self.target, line)
                if irclib.is_channel(self.target):
                    self.session.log.info("-%s- <%s> %s" % (self.target,
                                                            self.session.nick,
                                                            line))
                else:
                    self.session.log.info("<%s> %s" % (self.session.nick,
                                                       line))



#   Copyright 2014 Philip Schwartz
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Mirc Colors Unit Tests"""

import unittest

from pyhole.core import color
from pyhole.core.colormaps import mirccolors


class TestMircColors(unittest.TestCase):
    def setUp(self):
        self._colors = color.Color(mirccolors.MircColors)

    def test_colorize(self):
        result = self._colors.colorize("red", "blue", "test")
        self.assertEqual(result, u'\x034,2test\x0f')

    def test_colorize_fg(self):
        result = self._colors.colorize("red", text="test")
        self.assertEqual(result, u'\x034test\x0f')

    def test_colorize_bg(self):
        result = self._colors.colorize(bg="blue", text="test")
        self.assertEqual(result, u'\x03,2test\x0f')

    def test_long_colorize(self):
        enc = self._colors.colorize("red", text="test")
        result = enc + " Nice " + enc
        self.assertEqual(result, u'\x034test\x0f Nice \x034test\x0f')

    def test_clean(self):
        result = self._colors.clean(u'\x030,2test\x0f')
        self.assertEqual(result, u'test')

    def test_long_clean(self):
        result = self._colors.clean(u'\x030test\x0f Nice \x030test\x0f')
        self.assertEqual(result, u'test Nice test')



#   Copyright 2011-2016 Paul Voccio
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Launchpad Plugin"""

from launchpadlib.launchpad import Launchpad as LP

from pyhole.core import plugin
from pyhole.core import utils


class Launchpad(plugin.Plugin):
    """Provide access to the Launchpad API."""

    def __init__(self, session):
        self.session = session
        self.name = self.__class__.__name__
        self.launchpad = LP.login_anonymously("pyhole", "production",
                                              utils.get_directory(self.name))

    @plugin.hook_add_command("lbugs")
    @utils.require_params
    @utils.spawn
    def lbugs(self, message, params=None, **kwargs):
        """Launchpad bugs for a team (ex: .lbugs <project> <team>|<user>)."""
        project, team = params.split(" ", 2)
        try:
            members = self.launchpad.people[team]
            proj = self.launchpad.projects[project]

            if len(members.members) < 2:
                # Find a single member
                self._find_bugs(members, proj)
            else:
                # Find everyone on the team
                for i, person in enumerate(members.members):
                    if i <= 4:
                        self._find_bugs(message, person, proj, False)
                    else:
                        msg = "[...] truncated last %d users"
                        message.dispatch(msg % (len(members.members) - i))
                        break
        except KeyError:
            msg = "Unable to find user '%s' in Launchpad."
            message.dispatch(msg % team)

    @plugin.hook_add_keyword("lp")
    @utils.require_params
    @utils.spawn
    def keyword_lp(self, message, params=None, **kwargs):
        """Retrieve Launchpad bug information (ex: LP12345)."""
        params = utils.ensure_int(params)
        if not params:
            return

        try:
            bug = self.launchpad.bugs[params]
            task = bug.bug_tasks[len(bug.bug_tasks) - 1]
            message.dispatch("LP %s [Status: %s, Assignee: %s] %s" % (
                             task.title, task.status,
                             self._find_name(task.assignee_link),
                             bug.web_link))
        except Exception:
            return

    @plugin.hook_add_msg_regex("https?:\/\/bugs\.launchpad\.net\/.*\/\+bug")
    def regex_match_lp_bug_url(self, message, params=None, **kwargs):
        """Watch for Launchpad bug URLs."""
        try:
            line = message.message.split("/")
            for i, word in enumerate(line):
                if word == "+bug":
                    bug_id = line[i + 1].split(" ", 1)[0]
                    self.keyword_lp(bug_id)
        except TypeError:
            return

    @plugin.hook_add_msg_regex("https?:\/\/bugs\.launchpad\.net\/bugs")
    def regex_match_short_lp_bug_url(self, message, params=None, **kwargs):
        """Watch for short Launchpad bug URLs."""
        try:
            line = message.message.split("/")
            for i, word in enumerate(line):
                if word == "bugs":
                    bug_id = line[i + 1].split(" ", 1)[0]
                    self.keyword_lp(bug_id)
        except TypeError:
            return

    def _find_name(self, user):
        """Lookup a Launchpad user's display name."""
        try:
            return self.launchpad.people[user].display_name
        except ValueError:
            return "None"

    def _find_bugs(self, message, person, project, single=True):
        """Lookup Launchpad bugs."""
        bugs = project.searchTasks(assignee=person)
        for i, bug in enumerate(bugs):
            if i <= 4:
                message.dispatch("LP %s [Assignee: %s] %s" % (bug.title,
                                 person.display_name, bug.web_link))
            else:
                message.dispatch("[...] truncated last %d bugs" % (
                                 len(bugs) - i))
                break

        if single and i < 1:
            message.dispatch("No bugs found: %s" % (person.display_name))



#   Copyright 2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Slack Client Class"""

import slackclient
import time

from pyhole.core import log
from pyhole.core import plugin
from pyhole.core import utils
from pyhole.core import version
from pyhole.core.slack import message


CONFIG = utils.get_config()
LOG = log.get_logger()


class Client(object):
    """A Slack Connection"""

    def __init__(self, network):
        log.setup_logger(str(network))
        self.network_config = utils.get_config(network)

        self.addressed = False
        self.client = None
        self.log = log.get_logger(str(network))
        self.version = version.version_string()

        self.admins = CONFIG.get("admins", type="list")
        self.command_prefix = CONFIG.get("command_prefix")

        self.api_token = self.network_config.get("api_token")
        self.nick = self.network_config.get("nick")

        self.load_plugins()

    def load_plugins(self, reload_plugins=False):
        """Load plugins and their commands respectively."""
        if reload_plugins:
            plugin.reload_plugins(session=self)
        else:
            plugin.load_plugins(session=self)

        self.log.info("Loaded Plugins: %s" % plugin.active_plugins())
        plugin.run_hook_polls(self)

    def start(self):
        """Run the Slack network connection."""
        self.client = slackclient.SlackClient(self.api_token)
        self.client.rtm_connect()

        while True:
            try:
                for response in self.client.rtm_read():
                    self.log.debug(response)

                    if all(x in response for x in ("text", "channel", "user")):
                        r_channel = response["channel"]
                        r_user = response["user"]

                        try:
                            channel = self.client.server.channels.find(
                                r_channel).name
                            user = self.client.server.users.find(r_user).name
                        except AttributeError:
                            continue

                        msg = response["text"]

                        self.log.info("-#%s- <%s> %s" % (channel, user, msg))

                        _msg = message.Reply(self, msg, user, channel)
                        plugin.poll_messages(self, _msg)

                time.sleep(.1)
            except Exception:
                # NOTE(jk0): Disconnected? Try to reconnect.
                self.client.rtm_connect()
                continue

    def reply(self, target, msg):
        """Reply to a channel."""
        channel = self.client.server.channels.find(target)
        channel.send_message(msg)

    def op_user(self, *args, **kwargs):
        pass

    def deop_user(self, *args, **kwargs):
        pass

    def set_nick(self, *args, **kwargs):
        pass

    def join_channel(self, *args, **kwargs):
        pass

    def part_channel(self, *args, **kwargs):
        pass

    def privmsg(self, *args, **kwargs):
        pass



#   Copyright 2010-2016 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Administration Plugin"""

from pyhole.core import plugin
from pyhole.core import utils


class Admin(plugin.Plugin):
    """Administer your bot."""

    @plugin.hook_add_command("help")
    def help(self, message, params=None, **kwargs):
        """Learn how to use plugins (ex: .help <command>)."""

        if params:
            doc = _find_doc_string(params)
            if doc:
                message.dispatch(doc)
            else:
                message.dispatch("No help available: %s" % params)
        else:
            active_commands = plugin.active_commands()
            active_keywords = plugin.active_keywords()
            active_patterns = plugin.active_msg_regexs()
            active_pollers = plugin.active_polls()

            message.dispatch(self.help.__doc__)

            if len(active_commands) > 0:
                message.dispatch("Active Commands: %s" % active_commands)
            if len(active_keywords) > 0:
                message.dispatch("Active Keywords: %s" % active_keywords)
            if len(active_patterns) > 0:
                message.dispatch("Active Patterns: %s" % active_patterns)
            if len(active_pollers) > 0:
                message.dispatch("Active Pollers: %s" % active_pollers)

    @plugin.hook_add_command("version")
    def version(self, message, params=None, **kwargs):
        """Display the current version."""
        message.dispatch(self.session.version)

    @plugin.hook_add_command("reload")
    @utils.admin
    def reload(self, message, params=None, **kwargs):
        """Reload all plugins."""
        self.session.load_plugins(reload_plugins=True)
        message.dispatch("Loaded Plugins: %s" % plugin.active_plugins())

    @plugin.hook_add_command("op")
    @utils.admin
    @utils.require_params
    def op(self, message, params=None, **kwargs):
        """Op a user (ex: .op <channel> <nick>)."""
        self.session.op_user(params)

    @plugin.hook_add_command("deop")
    @utils.admin
    @utils.require_params
    def deop(self, message, params=None, **kwargs):
        """De-op a user (ex: .deop <channel> <nick>)."""
        self.session.deop_user(params)

    @plugin.hook_add_command("nick")
    @utils.admin
    @utils.require_params
    def nick(self, message, params=None, **kwargs):
        """Change nick (ex: .nick <nick>)."""
        self.session.set_nick(params)

    @plugin.hook_add_command("join")
    @utils.admin
    @utils.require_params
    def join(self, message, params=None, **kwargs):
        """Join a channel (ex: .join #channel [<key>])."""
        self.session.join_channel(params)

    @plugin.hook_add_command("part")
    @utils.admin
    @utils.require_params
    def part(self, message, params=None, **kwargs):
        """Part a channel (ex: .part <channel>)."""
        self.session.part_channel(params)

    @plugin.hook_add_command("say")
    @utils.admin
    @utils.require_params
    def say(self, message, params=None, **kwargs):
        """Send a PRIVMSG (ex: .say <channel>|<nick> message)."""
        (target, msg) = params.split(" ", 1)
        self.session.privmsg(target, msg)


def _find_doc_string(params):
    """Find the doc string for a plugin, command or keyword hook."""
    for p in plugin.active_plugin_classes():
        if p.__name__.upper() == params.upper():
            return p.__doc__

    for _, cmd_hook, cmd in plugin.hook_get_commands():
        if cmd.upper() == params.upper():
            return cmd_hook.__doc__

    for _, kw_hook, kw in plugin.hook_get_keywords():
        if kw.upper() == params.upper():
            return kw_hook.__doc__

    return None



#   Copyright 2010-2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""IRC Client Class"""

import random
import ssl
import time

import irc.client as irclib
from irc import connection

from pyhole.core import log
from pyhole.core import plugin
from pyhole.core import utils
from pyhole.core import version
from pyhole.core.irc import message


CONFIG = utils.get_config()
LOG = log.get_logger()


class Client(irclib.SimpleIRCClient):
    """An IRClib connection."""

    def __init__(self, network):
        irclib.SimpleIRCClient.__init__(self)
        log.setup_logger(str(network))
        self.network_config = utils.get_config(network)
        self.log = log.get_logger(str(network))
        self.version = version.version_string()
        self.source = None
        self.target = None
        self.addressed = False

        self.admins = CONFIG.get("admins", type="list")
        self.command_prefix = CONFIG.get("command_prefix")
        self.reconnect_delay = CONFIG.get("reconnect_delay", type="int")
        self.rejoin_delay = CONFIG.get("rejoin_delay", type="int")

        self.server = self.network_config.get("server")
        self.password = self.network_config.get("password", default=None)
        self.port = self.network_config.get("port", type="int", default=6667)
        self.ssl = self.network_config.get("ssl", type="bool", default=False)
        self.ipv6 = self.network_config.get("ipv6", type="bool", default=False)
        self.bind_to = self.network_config.get("bind_to", default=None)
        self.nick = self.network_config.get("nick")
        self.username = self.network_config.get("username", default=None)
        self.identify_password = self.network_config.get("identify_password",
                                                         default=None)
        self.channels = self.network_config.get("channels", type="list")

        self.load_plugins()

        self.log.info("Connecting to %s:%d as %s" % (self.server, self.port,
                      self.nick))
        self.connect(self.server, self.port, self.nick, self.password,
                     username=self.username,
                     connect_factory=connection.Factory(
                         wrapper=ssl.wrap_socket,
                         ipv6=self.ipv6).connect
                     if self.ssl else connection.Factory())

    def load_plugins(self, reload_plugins=False):
        """Load plugins and their commands respectively."""
        if reload_plugins:
            plugin.reload_plugins(session=self)
        else:
            plugin.load_plugins(session=self)

        self.log.info("Loaded Plugins: %s" % plugin.active_plugins())
        plugin.run_hook_polls(self)

    def notice(self, target, msg):
        """Send a notice."""
        self.connection.notice(target, msg)

    def privmsg(self, target, msg):
        """Send a privmsg."""
        self.connection.privmsg(target, msg)

    def op_user(self, params):
        """Op a user."""
        params = params.split(" ", 1)
        self.connection.mode(params[0], "+o %s" % params[1])

    def deop_user(self, params):
        """De-op a user."""
        params = params.split(" ", 1)
        self.connection.mode(params[0], "-o %s" % params[1])

    def set_nick(self, params):
        """Set IRC nick."""
        self.nick = params
        self.connection.nick(params)

    def join_channel(self, params):
        """Join a channel."""
        channel = params.split(" ", 1)
        self.reply("Joining %s" % channel[0])
        if irclib.is_channel(channel[0]):
            self.channels.append(channel[0])
            if len(channel) > 1:
                self.connection.join(channel[0], channel[1])
            else:
                self.connection.join(channel[0])

    def part_channel(self, params):
        """Part a channel."""
        self.channels.remove(params)
        self.reply("Parting %s" % params)
        self.connection.part(params)

    def on_nicknameinuse(self, connection, _event):
        """Ensure the use of unique IRC nick."""
        random_int = random.randint(1, 100)
        self.log.info("IRC nick '%s' is currently in use" % self.nick)
        self.nick = "%s%d" % (self.nick, random_int)
        self.log.info("Setting IRC nick to '%s'" % self.nick)
        connection.nick("%s" % self.nick)
        # Try to prevent nick flooding
        time.sleep(1)

    def on_welcome(self, connection, _event):
        """Join channels upon successful connection."""
        if self.identify_password:
            self.privmsg("NickServ", "IDENTIFY %s" % self.identify_password)

        for channel in self.channels:
            channel = channel.split(" ", 1)
            if irclib.is_channel(channel[0]):
                if len(channel) > 1:
                    connection.join(channel[0], channel[1])
                else:
                    connection.join(channel[0])

    def on_disconnect(self, _connection, _event):
        """Attempt to reconnect after disconnection."""
        self.log.info("Disconnected from %s:%d" % (self.server, self.port))
        self.log.info("Reconnecting in %d seconds" % self.reconnect_delay)
        time.sleep(self.reconnect_delay)
        self.log.info("Connecting to %s:%d as %s" % (self.server, self.port,
                      self.nick))
        self.connect(self.server, self.port, self.nick, self.password,
                     username=self.username,
                     connect_factory=connection.Factory(
                         wrapper=ssl.wrap_socket,
                         ipv6=self.ipv6).connect
                     if self.ssl else connection.Factory())

    def on_kick(self, connection, event):
        """Automatically rejoin channel if kicked."""
        source = event.source.nick
        target = event.target
        nick, reason = event.arguments

        if nick == self.nick:
            self.log.info("-%s- kicked by %s: %s" % (target, source, reason))
            self.log.info("-%s- rejoining in %d seconds" % (target,
                          self.rejoin_delay))
            time.sleep(self.rejoin_delay)
            connection.join(target)
        else:
            self.log.info("-%s- %s was kicked by %s: %s" % (target, nick,
                          source, reason))

    def on_invite(self, _connection, event):
        """Join a channel upon invitation."""
        source = event.source.split("@", 1)[0]
        if source in self.admins:
            self.join_channel(event.arguments[0])

    def on_ctcp(self, connection, event):
        """Respond to CTCP events."""
        source = event.source.nick
        ctcp = event.arguments[0]

        if ctcp == "VERSION":
            self.log.info("Received CTCP VERSION from %s" % source)
            connection.ctcp_reply(source, "VERSION %s" % self.version)
        elif ctcp == "PING":
            if len(event.arguments) > 1:
                self.log.info("Received CTCP PING from %s" % source)
                connection.ctcp_reply(source, "PING %s" % event.arguments[1])

    def on_join(self, _connection, event):
        """Handle joins."""
        target = event.target
        source = event.source.nick
        self.log.info("-%s- %s joined" % (target, source))

    def on_part(self, _connection, event):
        """Handle parts."""
        target = event.target
        source = event.source.nick
        self.log.info("-%s- %s left" % (target, source))

    def on_quit(self, _connection, event):
        """Handle quits."""
        source = event.source.nick
        self.log.info("%s quit" % source)

    def on_action(self, _connection, event):
        """Handle IRC actions."""
        target = event.target
        source = event.source.nick
        msg = event.arguments[0]
        self.log.info("-%s- * %s %s" % (target, source, msg))

    def on_privnotice(self, _connection, event):
        """Handle private notices."""
        if event.source is not None:
            source = event.source.nick
        else:
            source = None
        msg = event.arguments[0]
        self.log.info("-%s- %s" % (source, msg))

    def on_pubnotice(self, _connection, event):
        """Handle public notices."""
        target = event.target
        if event.source is not None:
            source = event.source.nick
        else:
            source = None
        msg = event.arguments[0]
        self.log.info("-%s- <%s> %s" % (target, source, msg))

    def on_privmsg(self, _connection, event):
        """Handle private messages."""
        msg = event.arguments[0]

        source = event.source.split("@", 1)[0]
        target = event.source.nick
        _msg = message.Reply(self, msg, source, target)

        if self.target != self.nick:
            self.log.info("<%s> %s" % (target, msg))
            plugin.poll_messages(self, _msg, private=True)

    def on_pubmsg(self, _connection, event):
        """Handle public messages."""
        nick = event.source.nick
        msg = event.arguments[0]

        source = event.source.split("@", 1)[0]
        target = event.target

        self.log.info("-%s- <%s> %s" % (target, nick, msg))

        _msg = message.Reply(self, msg, source, target)
        plugin.poll_messages(self, _msg)



#   Copyright 2011-2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import setuptools

from pyhole.core import version


setuptools.setup(
    name="irc-pyhole",
    version=version.version(),
    author="Josh Kearney",
    author_email="josh@jk0.org",
    description="A modular IRC & Slack bot.",
    url="https://github.com/jk0/pyhole",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python"
    ],
    install_requires=[
        "BeautifulSoup==3.2.0",
        "Eventlet",
        "irc",
        "launchpadlib",
        "pywunderground",
        "requests",
        "slackclient",
        "yahoo-finance"
    ],
    setup_requires=[
        "flake8==2.4.1",
        "sphinx"
    ],
    test_suite="pyhole.tests",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "pyhole = pyhole.main:Main"
        ]
    }
)



#   Copyright 2015 Jason Meridth
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Allergies Plugin"""

import datetime

from pyhole.core import plugin
from pyhole.core import utils


class Allergies(plugin.Plugin):
    """Provide access to current allergy data."""

    @plugin.hook_add_command("allergies")
    @utils.spawn
    def allergies(self, message, params=None, **kwargs):
        """Display current allergies in San Antonio, TX (ex: .allergies)."""
        d = datetime.datetime.now()
        weekend = d.isoweekday() in (6, 7)
        if weekend:
            message.dispatch("Unable to fetch allergy data on weekends.")
            return

        today = d.strftime("%Y-%m-%d")
        url = "http://saallergy.info/day/%s" % today
        headers = {"accept": "application/json"}

        response = utils.fetch_url(url, headers=headers)
        if response.status_code != 200:
                return

        data = response.json()
        text = "Allergies for %s: " % today
        for a in data["results"]:
            text = text + "%s - %s (%s) | " % (a["allergen"], a["level"],
                                               a["count"])
        text = text.rstrip(" ")
        text = text.rstrip("|")

        message.dispatch(text)

    @plugin.hook_add_command("pollen")
    def alias_pollen(self, message, params=None, **kwargs):
        """Alias of allergies."""
        self.allergies(message, params, **kwargs)



#   Copyright 2011-2015 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Config Unit Tests"""

import unittest

from pyhole.core import config
from pyhole.core import utils


class TestConfig(unittest.TestCase):
    def setUp(self):
        home_dir = utils.get_home_directory()
        self.config = config.Config(home_dir + "pyhole.conf", "Pyhole")

    def test_missing_parameter(self):
        self.assertRaises(SystemExit, self.config.get, "foo")

    def test_sections(self):
        sections = self.config.sections()
        self.assertTrue(isinstance(sections, list))
        self.assertTrue(len(sections) > 1)

    def test_get_int(self):
        test_int = self.config.get("reconnect_delay", type="int")
        self.assertTrue(isinstance(test_int, int))

    def test_get_bool(self):
        test_bool = self.config.get("debug", type="bool")
        self.assertTrue(isinstance(test_bool, bool))

    def test_get_list(self):
        test_list = self.config.get("admins", type="list")
        self.assertTrue(isinstance(test_list, list))

    def test_get_str(self):
        test_str = self.config.get("command_prefix")
        self.assertTrue(isinstance(test_str, str))



#   Copyright 2010-2016 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Redmine Plugin"""

import json

from pyhole.core import plugin
from pyhole.core import utils


class Redmine(plugin.Plugin):
    """Provide access to the Redmine API."""

    def __init__(self, session):
        self.session = session
        self.name = self.__class__.__name__

        self.redmine = utils.get_config("Redmine")
        self.redmine_domain = self.redmine.get("domain")
        self.redmine_key = self.redmine.get("key")
        self.redmine_url = "https://%s:password@%s" % (
            self.redmine_key, self.redmine_domain)

    @plugin.hook_add_command("rbugs")
    @utils.require_params
    @utils.spawn
    def rbugs(self, message, params=None, **kwargs):
        """Redmine bugs for a user (ex: .rbugs <login>)."""
        login = params.split(" ", 1)[0]
        user_id = self._find_user(login)

        i = 0
        issues = self._find_issues(user_id)
        for i, issue in enumerate(issues):
            if i <= 4:
                self._find_issue(message, issue["id"])
            else:
                message.dispatch("[...] truncated last %d bugs" % (
                                 len(issues) - i))
                break
        else:
            if i <= 0:
                message.dispatch("No Redmine bugs found: '%s'" % login)

    @plugin.hook_add_keyword("rm")
    @utils.require_params
    @utils.spawn
    def keyword_rm(self, message, params=None, **kwargs):
        """Retrieve Redmine bug information (ex: RM12345)."""
        params = utils.ensure_int(params)
        if params:
            self._find_issue(message, params)

    @plugin.hook_add_msg_regex("https?:\/\/redmine\..*/issues")
    def regex_match_rm_bug_url(self, message, params=None, **kwargs):
        """Watch for Redmine bug URLs."""
        try:
            line = message.message.split("/")
            for i, word in enumerate(line):
                if word == "issues":
                    bug_id = line[i + 1].split(" ", 1)[0]
                    self.keyword_rm(bug_id)
        except TypeError:
            return

    def _find_issues(self, user_id):
        """Find all issues for a Redmine user."""
        url = "%s/issues.json?assigned_to_id=%s" % (
              self.redmine_url, user_id)
        response = utils.fetch_url(url)
        if response.status_code != 200:
                return

        return json.loads(response.content)["issues"]

    def _find_user(self, login):
        """Find a specific Redmine user."""
        for user in self._find_users():
            if login == user["login"]:
                return user["id"]

        for user in self._find_users(100):
            if login == user["login"]:
                return user["id"]

    def _find_users(self, offset=None):
        """Find all Redmine users."""
        if offset:
            url = "%s/users.json?limit=100&offset=%d" % (
                  self.redmine_url, offset)
        else:
            url = "%s/users.json?limit=100" % self.redmine_url
        response = utils.fetch_url(url)
        if response.status_code != 200:
                return

        return json.loads(response.content)["users"]

    def _find_issue(self, message, issue_id):
        """Find and display a Redmine issue."""
        url = "%s/issues/%s.json" % (self.redmine_url, issue_id)
        response = utils.fetch_url(url)
        if response.status_code != 200:
                return

        try:
            issue = json.loads(response.content)["issue"]
        except Exception:
            return

        msg = "RM %s #%s: %s [Status: %s, Priority: %s, Assignee: %s]" % (
            issue["tracker"]["name"],
            issue["id"],
            issue["subject"],
            issue["status"]["name"],
            issue["priority"]["name"],
            issue.get("assigned_to", {}).get("name", "N/A"))
        url = "https://%s/issues/%s" % (self.redmine_domain, issue["id"])

        message.dispatch("%s %s" % (msg, url))



#   Copyright 2015-2016 Rick Harris
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Box Office Plugin"""

from BeautifulSoup import BeautifulSoup

from pyhole.core import plugin
from pyhole.core import utils


class Boxoffice(plugin.Plugin):
    """Display top box office movies."""

    @plugin.hook_add_command("boxoffice")
    @utils.spawn
    def boxoffice(self, message, params=None, **kwargs):
        """Display top box office movies"""
        resp = utils.fetch_url("http://www.rottentomatoes.com")
        if resp.status_code != 200:
            message.dispatch("Could not fetch box office results.")
            return

        message.dispatch("Top 10 at the Box Office")
        message.dispatch("=" * 64)
        soup = BeautifulSoup(resp.text)
        rows = soup.findAll("table", id="Top-Box-Office")[0].findAll("tr")
        for row in rows:
            children = row.findAll("td")
            try:
                score = row("span", **{"class": "tMeterScore"})[0].contents[0]
                title = children[1].find("a").string
                take = children[2].string.strip()
            except Exception:
                pass
            msg = "%s %s%s" % (score.rjust(3), title.ljust(40), take.rjust(10))
            message.dispatch(msg)
